// -*- Mode: c++ -*-
// $Id: grid3d.cpp,v 1.11 2000/08/08 21:28:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/grid3d.cpp,v $

#if defined(_WIN32) || defined(__sgi)
#include <float.h>
#endif
#include "options.h"
#include "grid3d.h"

// place a triangle mesh in the grid
void Grid3D::placeIntoGrid(TMesh *mesh) {
  assert(mesh != 0);

  _min = cvec3d(DBL_MAX, DBL_MAX, DBL_MAX);
  _max = cvec3d(-DBL_MAX, -DBL_MAX, -DBL_MAX);

  // compute bounding box
  int totalTriangles = 0;

  TMesh::FaceIterType ti;
  for (ti = mesh->faceBegin(); ti != mesh->faceEnd(); ++ti) {
    if ((*ti)->isLeaf()) {
      totalTriangles++;
      for (int i=0; i<(*ti)->noVtx(); i++) {
	_min = _min.min((*ti)->vert(i)->limitPos());
	_max = _max.max((*ti)->vert(i)->limitPos());
      }
    }
  }

  cerr << totalTriangles << " triangles ... ";

  // select grid size
  int grid_size;
  if (totalTriangles <= 10000) grid_size = 10;
  else if (totalTriangles <= 40000) grid_size = 15;
  else grid_size = 20;

  // create grid
  createGrid(grid_size);

  _xstep = (_max.x() - _min.x()) / _size;
  _ystep = (_max.y() - _min.y()) / _size;
  _zstep = (_max.z() - _min.z()) / _size;

  // put triangles into grid
  for (ti = mesh->faceBegin(); ti != mesh->faceEnd(); ++ti) {
    if ((*ti)->isLeaf()) {
      placeTriangle(*ti);
    }
  }
}

// allocate proper space for a grid of size s, free old
void Grid3D::createGrid(int s) {
  int i, j, k;
  // clear old space
  if (_grid && (_size > 0)) {
    for (i=0; i<_size; i++)
      for (j=0; j<_size; j++)
	for (k=0; k<_size; k++)
	  _grid[i][j][k].clear();
  }

  // if size changed, free old and allocate new
  if (s != _size) {
    if (_grid) {
      for (i=0; i<_size; i++)
	for (j=0; j<_size; j++)
	  delete[] _grid[i][j];
      for (i=0; i<_size; i++)
	delete[] _grid[i];
      delete[] _grid;
    }
    _size = s;
    _grid = new vector<GridTri*>**[s];
    for (i=0; i<s; i++)
      _grid[i] = new vector<GridTri*>*[s];
    for (i=0; i<s; i++)
      for (j=0; j<s; j++)
	_grid[i][j] = new vector<GridTri*>[s];
  }
}

// free space
Grid3D::~Grid3D() {
  int i, j, k;
  if (_grid && (_size > 0)) {
    for (i=0; i<_size; i++) {
      for (j=0; j<_size; j++) {
	for (k=0; k<_size; k++) {
	  _grid[i][j][k].clear();
	}
	delete[] _grid[i][j];
      }
      delete[] _grid[i];
    }
    delete[] _grid;
  }
  _grid = 0;
  _rays.clear();
}

// place a triangle into intesected cells
void Grid3D::placeTriangle(Tri *t) {
  // bounding box
  cvec3d tmin(DBL_MAX, DBL_MAX, DBL_MAX), tmax(-DBL_MAX, -DBL_MAX, -DBL_MAX);
  for (int v=0; v<t->noVtx(); v++) {
    tmin = tmin.min(t->vert(v)->limitPos());
    tmax = tmax.max(t->vert(v)->limitPos());
  }
  // starting and ending cells of the bounding box
  CVec3T<int> start = 
    CVec3T<int>(::max((int)floor((tmin.x() - _min.x())/_xstep), 0),
		::max((int)floor((tmin.y() - _min.y())/_ystep), 0),
		::max((int)floor((tmin.z() - _min.z())/_zstep), 0)),
    end =
    CVec3T<int>(::min((int)floor((tmax.x() - _min.x())/_xstep), _size-1),
		::min((int)floor((tmax.y() - _min.y())/_ystep), _size-1),
		::min((int)floor((tmax.z() - _min.z())/_zstep), _size-1));

  GridTri* gt = new GridTri(t);

  bool sa = (gt->n.x() >= 0), sb = (gt->n.y() >= 0), sc = (gt->n.z() >= 0);
  cvec3d p1, p2;
  bool sd1, sd2;

  // put triangle in the cells in its bounding box
  // still excessive, treat triangle as a plane to detect intersection
  for (int i=start.x(); i<=end.x(); i++) {
    p1.x() = _min.x() + i*_xstep;
    p2.x() = p1.x() + _xstep;
    for (int j=start.y(); j<=end.y(); j++) {
      if (sb == sa) {
	p1.y() = _min.y() + j*_ystep; p2.y() = p1.y() + _ystep;
      } else {
	p2.y() = _min.y() + j*_ystep; p1.y() = p2.y() + _ystep;
      }
      for (int k=start.z(); k<=end.z(); k++) {
	if (sc == sa) {
	  p1.z() = _min.z() + k*_zstep; p2.z() = p1.z() + _zstep;
	} else {
	  p2.z() = _min.z() + k*_zstep; p1.z() = p2.z() + _zstep;
	}
	sd1 = gt->n.dot(p1 - gt->v(0)) >= 0;
	sd2 = gt->n.dot(p2 - gt->v(0)) >= 0;
	if (sd1 != sd2) _grid[i][j][k].push_back(gt);
      }
    }
  }
}

// intersect a ray with the mesh
bool Grid3D::intersectRay(const cvec3d& start, const cvec3d& end) {
  // pick rays
  _rays.push_back(Ray(start, end));
  cvec3d dir = start - end;

  int idir[3] = {
    fabs(dir.x()) < DBL_EPSILON ? 0 : (dir.x() > 0 ? 1 : -1),
    fabs(dir.y()) < DBL_EPSILON ? 0 : (dir.y() > 0 ? 1 : -1),
    fabs(dir.z()) < DBL_EPSILON ? 0 : (dir.z() > 0 ? 1 : -1) };

  double dist = 0;
  cvec3d pos = end - _min;

  int ix = (int)floor((end.x() - _min.x())/_xstep),
    iy = (int)floor((end.y() - _min.y())/_ystep),
    iz = (int)floor((end.z() - _min.z())/_zstep);
  double tx = 1.0, ty = 1.0, tz = 1.0;

  // first cell
  ix = ::min(::max(ix, 0), _size-1);
  iy = ::min(::max(iy, 0), _size-1);
  iz = ::min(::max(iz, 0), _size-1);

  // intersection test, from end to start
  while ((dist < 1) && (ix >= 0) && (ix < _size) &&
	 (iy >= 0) && (iy < _size) && (iz >= 0) && (iz < _size)) {
    int csz = _grid[ix][iy][iz].size();
    for (int i=0; i<csz; i++) {
      GridTri *gt = _grid[ix][iy][iz][i];
      double rpdot = dir.dot(gt->n);
      if (rpdot != 0) {
        double t = (gt->v(0) - end).dot(gt->n) / rpdot;
	if (t > 0.000001 && t < 1) {
	  cvec3d pt = end + t*dir;
          if (((pt - gt->v(0)).dot(gt->en1) > - Options::EPSILON) &&
              ((pt - gt->v(1)).dot(gt->en2) > - Options::EPSILON) &&
              ((pt - gt->v(2)).dot(gt->en3) > - Options::EPSILON)) {
	    _rays[_rays.size()-1].p = t;
	    return true;
	  }
	}
      }
    }

    // next cell
    if (idir[0] != 0)
      tx = (_min.x() + (ix+(idir[0]+1)/2)*_xstep - end.x()) / dir.x();
    if (idir[1] != 0)
      ty = (_min.y() + (iy+(idir[1]+1)/2)*_ystep - end.y()) / dir.y();
    if (idir[2] != 0)
      tz = (_min.z() + (iz+(idir[2]+1)/2)*_zstep - end.z()) / dir.z();
    if ((tx <= ty) && (tx <= tz)) {
      dist = tx, ix += idir[0];
      if (ty == tx) iy += idir[1];
      if (tz == tx) iz += idir[2];
    } else if (ty <= tz) {
      dist = ty, iy += idir[1];
      if (tz == ty) iz += idir[2];
    } else dist = tz, iz += idir[2];
  }

  return false;
}
