// -*- Mode: c++ -*-
// $Id: meshviewer.cpp,v 1.17 2000/08/09 19:45:51 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/meshviewer.cpp,v $

#include <string.h>
#include <GL/glut.h>
#include "options.h"
#include "meshviewer.h"
#include "silhouette.h"

MeshViewer::MeshViewer(const Camera &c, ArcBall *b, Grid3D *g, Silhouettes *s,
		       int width, int height)
  : Viewer("Mesh", width, height),
    _camera(c), _arcball(b), _grid(g), _mesh(0), _sil(s),
    _zooming(false), _translating(false),
    _renderList(-1), _uiAction(0) {
  setLightAndMaterial();
}

MeshViewer::~MeshViewer() {
  if ( glIsList(_renderList) ) glDeleteLists( _renderList, 1 );
  _renderList = -1;
  if (_uiAction) delete _uiAction;
  _uiAction = 0;
}

void MeshViewer::clearDisplayList() {
  if ( glIsList(_renderList) ) glDeleteLists( _renderList, 1 );
  _renderList = -1;
}

void MeshViewer::display() {
  _camera.loadMatrices();

  glColor3f(0.5f, 0.5f, 0.5f);
  render();
}

void MeshViewer::render() {
  if (Options::lighting)
    glEnable(GL_LIGHTING);
  else
    glDisable(GL_LIGHTING);

  if (Options::lighting)
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  if ( glIsList(_renderList) ) {
    glCallList( _renderList );
  } else {
    _renderList = glGenLists(1);
    glNewList( _renderList, GL_COMPILE_AND_EXECUTE );

    glLineWidth(2);
    glBegin(GL_TRIANGLES);
    TMesh::FaceIterType it;
    for(it = _mesh->faceBegin(); it != _mesh->faceEnd(); ++it) {
      if((*it)->isLeaf()) {
	EnoType e = (*it)->directEno(1, CCW);
	for(VnoType v = 0; v < 3; ++v, e = (*it)->nextEno(e)) {
	  glNormal3dv((*it)->normal((*it)->headVno(e)));
	  glVertex3dv(((*it)->headVert(e))->limitPos());
	}
      }
    }
    glEnd();

    glDisable(GL_LIGHTING);

    // draw wireframe
    if (Options::drawWireframe) {
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glDepthRange(0, 1 - depthOffset/2);
      glColor3ub(127, 127, 127);

      glLineWidth(2);
      glBegin(GL_TRIANGLES);
      TMesh::FaceIterType it;
      for(it = _mesh->faceBegin(); it != _mesh->faceEnd(); ++it) {
	if((*it)->isLeaf()) {
	  EnoType e = (*it)->directEno(1, CCW);
	  for(VnoType v = 0; v < 3; ++v, e = (*it)->nextEno(e)) {
	    glNormal3dv((*it)->normal((*it)->headVno(e)));
	    glVertex3dv(((*it)->headVert(e))->limitPos());
	  }
	}
      }
      glEnd();
      glDepthRange(0, 1);
    }

    // draw face normal
    if (Options::drawFaceNormal) {
      glLineWidth(1);
      glColor3ub(200, 0, 200);
      glBegin(GL_LINES);
      for(it = _mesh->faceBegin(); it != _mesh->faceEnd(); ++it) {
	if((*it)->isLeaf()) {
	  cvec3d center = 1.0/3*((*it)->vert(0)->limitPos()+
				  (*it)->vert(1)->limitPos()+
				  (*it)->vert(2)->limitPos());
	  glVertex3dv(center);
	  glVertex3dv(center + 0.2*(*it)->normal());
	}
      }
      glEnd();
    }

    glEndList();
  }

  // draw silhouette edges
  if (Options::drawSilhouetteEdges) {
    glLineWidth(3);
    glDepthRange(0, 1 - depthOffset);
    glColor3ub(0, 0, 255);
    glBegin(GL_LINES);
    int esz = _sil->edges().size();
    for (int i=0; i<esz; i++) {
      SilEdge *se = _sil->edges()[i];
      if (! se->isBoundary()) {
	Tri *tri = se->tri(0);
	EnoType eno = se->eno(0);
	glVertex3dv(tri->headVert(eno)->limitPos());
	glVertex3dv(tri->tailVert(eno)->limitPos());
      }
    }
    glEnd();
    glDepthRange(0, 1);
  }

  // draw boundary
  if (Options::drawBoundary && OcTree::boundary.size() > 0) {
    glLineWidth(3);
    glDepthRange(0, 1 - depthOffset);
    glColor3ub(0, 255, 0);
    glBegin(GL_LINES);
    int bsz = OcTree::boundary.size();
    for (int i=0; i<bsz; i++) {
      Tri *tri = OcTree::boundary[i].tri();
      EnoType eno = OcTree::boundary[i].eno();
      glVertex3dv(tri->headVert(eno)->limitPos());
      glVertex3dv(tri->tailVert(eno)->limitPos());
    }
    glEnd();
    glDepthRange(0, 1);
  }

  // draw picked rays
  if (Options::drawRays) {
    glLineWidth(2);
    glColor3ub(255, 0, 0);
    glBegin(GL_LINES);
    int rsz = _grid->rays().size();
    for (int i=0; i<rsz; i++) {
      glVertex3dv(_grid->rays()[i].start);
      glVertex3dv(_grid->rays()[i].end);
    }
    glEnd();
    glPointSize(3);
    glBegin(GL_POINTS);
    for (int j=0; j<rsz; j++) {
      glVertex3dv(_grid->rays()[j].end);
      if (_grid->rays()[j].p > 0)
	glVertex3dv(_grid->rays()[j].p * _grid->rays()[j].start +
		    (1-_grid->rays()[j].p) * _grid->rays()[j].end);
    }
    glEnd();
  }

  // draw grid
  if (Options::drawMeshGrid) {
    cvec3d s = _grid->maxp() - _grid->minp();
    glLineWidth(1);
    glColor3ub(0, 200, 0);
    glBegin(GL_LINES);
    int gsz = _grid->size();
    for (int i=0; i<=gsz; i++) {
      for (int j=0; j<=gsz; j++) {
	glVertex3dv(_grid->minp()+cvec3d(0, i*_grid->ystep(), j*_grid->zstep()));
	glVertex3dv(_grid->minp()+cvec3d(s.x(), i*_grid->ystep(), j*_grid->zstep()));
	glVertex3dv(_grid->minp()+cvec3d(i*_grid->xstep(), 0, j*_grid->zstep()));
	glVertex3dv(_grid->minp()+cvec3d(i*_grid->xstep(), s.y(), j*_grid->zstep()));
	glVertex3dv(_grid->minp()+cvec3d(i*_grid->xstep(), j*_grid->ystep(), 0));
	glVertex3dv(_grid->minp()+cvec3d(i*_grid->xstep(), j*_grid->ystep(), s.z()));
      }
    }
    glEnd();
    glColor3ub(255, 0, 0);
    glPointSize(3);
    glBegin(GL_POINTS);
    glVertex3dv(_grid->minp());
    glEnd();
  }
}

void MeshViewer::reshape(int w, int h) {
  if ((w == 0) || (h == 0) || ((w == _width) && (h == _height))) return;
  _width = w;
  _height = h;
  glViewport(0, 0, w, h);

  int vp[] = {0, 0, w, h};
  _camera.setViewport(vp);
  _camera.computeProjection();
  redisplay();
}

void MeshViewer::mouse(int button, int state, int x, int y) {
  _zooming = false;
  _translating = false;
  if (_uiAction) {
    delete _uiAction;
    _uiAction = 0;
  }
  y = _height - y; 

  if (state == GLUT_DOWN) {
    switch(button) {
    case GLUT_LEFT_BUTTON: 
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_translating = true;
	_transX = x, _transY = y;
      } else {
	_uiAction = new CameraRotateAction(x, y, _arcball, &_camera); 
      }
      break;
    case GLUT_MIDDLE_BUTTON: 
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_uiAction = new CameraTransXYAction(x, y, _arcball, &_camera);
      }
      // translating, not touching camera modelview
      else {
	_translating = true;
	_transX = x, _transY = y;
      }
      break;
    case GLUT_RIGHT_BUTTON: 
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
	_uiAction = new CameraTransZAction(x, y, _arcball, &_camera);
      }
      // zooming, not touching camera modelview
      else {
	_zooming = true;
	_zoomY = y;
      }
      break;
    }
  }
  redisplay();
}

void MeshViewer::motion(int x, int y) {
  y = _height - y; 
  if (_zooming) {
    _camera.scaleWindow(2*double(_zoomY - y)/_height);
    _zoomY = y;
    redisplay();
  } else if (_translating) {
    _camera.translateWindow(double(x-_transX)/_width,
			    double(y-_transY)/_height);
    _transX = x;
    _transY = y;
    redisplay();
  } else if(_uiAction) {
    _uiAction->update(x, y);
    redisplay();
  }
}

void MeshViewer::setLight() {
  // light
  GLfloat lgt1_diffuse[] =  { 0.5f, 0.5f, 0.5f, 1.0f };
  GLfloat lgt1_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
  GLfloat lgt1_ambient[] =  {0.1f, 0.1f, 0.1f, 1.0f};
  
  GLfloat lgt2_diffuse[] =  { 0.5f, 0.5f, 0.5f, 1.0f };
  GLfloat lgt2_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
  GLfloat lgt2_ambient[] =  {0.1f, 0.1f, 0.1f, 1.0f};

  GLfloat light_pos0[] =   { 1.0f,  1.0f, 1.0f, 0.0f };
  GLfloat light_pos1[] =   { 1.0f,  1.0f, 0.0f, 0.0f };
  GLfloat light_pos2[] =   { -1.0f, 1.0f, 0.0f, 0.0f };
  
  glLightfv(GL_LIGHT0, GL_POSITION,light_pos0);
  
  glLightfv(GL_LIGHT1, GL_POSITION,light_pos1);
  glLightfv(GL_LIGHT1, GL_SPECULAR,lgt1_specular);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,lgt1_diffuse);
  glLightfv(GL_LIGHT1, GL_AMBIENT, lgt1_ambient);
  
  glLightfv(GL_LIGHT2, GL_POSITION,light_pos2);
  glLightfv(GL_LIGHT2, GL_SPECULAR,lgt2_specular);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lgt2_diffuse);
  glLightfv(GL_LIGHT2, GL_AMBIENT, lgt2_ambient);
  
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT0);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);

  glEnable(GL_DEPTH_TEST);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void MeshViewer::setLightAndMaterial() {
  // Light
  setLight();

  // Material properties
  float matSpecular[] = {0.478814, 0.457627, 0.5};
  float matAmbient[] =  {0.75, 0.652647, 0.154303};
  float matDiffuse[] =  {0.75, 0.652647, 0.154303};
  float matFrontShininess = 25.0;

  float matBackSpecular[] = {0.1596,   0.1525,   0.1667};
  float matBackAmbient[] =  {0.3750,   0.3263,   0.0772};
  float matBackDiffuse[] =  {0.3750,   0.3263,   0.0772};
  float matBackShininess = 100.0;

  glMaterialfv(GL_FRONT, GL_DIFFUSE, matDiffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matSpecular);
  glMaterialfv(GL_FRONT, GL_AMBIENT, matAmbient);

  glMaterialfv(GL_BACK, GL_DIFFUSE, matBackDiffuse);
  glMaterialfv(GL_BACK, GL_SPECULAR, matBackSpecular);
  glMaterialfv(GL_BACK, GL_AMBIENT, matBackAmbient);
  
  glMaterialf(GL_FRONT, GL_SHININESS, matFrontShininess);
  glMaterialf(GL_BACK, GL_SHININESS, matBackShininess);
}
