// -*- Mode: c++ -*-
// $Id: silhouette.cpp,v 1.15 2000/08/08 21:28:16 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/src/silhouette.cpp,v $

#ifdef __GCC__
#include <cmath>
#else
#include <math.h>
#endif
#if defined(_WIN32) || defined(__sgi)
#include <float.h>
#endif
#include "options.h"
#include "silhouette.h"

// memory allocation constants
enum { chunk_size = 1000, chain_chunk_size = 100 };

// initial position _mem[_curr_chunk][_curr] should be the first empty object
int SilEdge::_curr = chunk_size;
int SilEdge::_curr_chunk = -1;
vector<SilEdge*> SilEdge::_mem;

int Intersection::_curr = chunk_size;
int Intersection::_curr_chunk = -1;
vector<Intersection*> Intersection::_mem;

int Chain::_curr = chain_chunk_size;
int Chain::_curr_chunk = -1;
vector<Chain*> Chain::_mem;

void* SilEdge::operator new(size_t s) {
  if (_curr >= chunk_size) {
    _curr = 0;
    _curr_chunk++;
    if (_curr_chunk >= _mem.size()) {
      SilEdge *se = ::new SilEdge[chunk_size];
      _mem.push_back(se);
    }
  }
  return &_mem[_curr_chunk][_curr++];
}

void SilEdge::clear() {
  _curr = chunk_size;
  _curr_chunk = -1;
}

void* Intersection::operator new(size_t s) {
  if (_curr >= chunk_size) {
    _curr = 0;
    _curr_chunk++;
    if (_curr_chunk >= _mem.size()) {
      Intersection *pt = ::new Intersection[chunk_size];
      _mem.push_back(pt);
    }
  }
  return &_mem[_curr_chunk][_curr++];
}

void Intersection::clear() {
  _curr = chunk_size;
  _curr_chunk = -1;
}

void* Chain::operator new(size_t s) {
  if (_curr >= chain_chunk_size) {
    _curr = 0;
    _curr_chunk++;
    if (_curr_chunk >= _mem.size()) {
      Chain *ch = ::new Chain[chain_chunk_size];
      _mem.push_back(ch);
    }
  }
  return &_mem[_curr_chunk][_curr++];
}

void Chain::clear() {
  _curr = chain_chunk_size;
  _curr_chunk = -1;
}

SilEdge::SilEdge( DualEdge *e, Camera *vc )
  : _onChain(false), _cluster(-1), _de(e), _be(0) {
  _v[0] = e->tri(0)->headVert(e->eno(0));
  _v[1] = e->tri(0)->tailVert(e->eno(0));
  // positions in the image plane
  _iv[0] = vc->project(_v[0]->limitPos());
  _iv[1] = vc->project(_v[1]->limitPos());
}

SilEdge::SilEdge( BoundaryEdge *e, Camera *vc )
  : _onChain(false), _cluster(-1), _de(0), _be(e) {
  _v[0] = e->v1();
  _v[1] = e->v2();
  // positions in the image plane
  _iv[0] = vc->project(_v[0]->limitPos());
  _iv[1] = vc->project(_v[1]->limitPos());
}

SilEdge::~SilEdge() {
  int sz = _intersection.size();
  for (int i=0; i<sz; i++)
    delete _intersection[i];
  _intersection.clear();
  _intersected.clear();
}

void SilEdge::clusterize( int c ) {
  if (_cluster == c) return;
  _cluster = c;
  vector<SilEdge*>::iterator ei;
  for (ei = v(0)->silEdges().begin(); ei != v(0)->silEdges().end(); ei++) {
    if ((*ei)->_cluster != c) (*ei)->clusterize(c);
  }
  for (ei = v(1)->silEdges().begin(); ei != v(1)->silEdges().end(); ei++) {
    if ((*ei)->_cluster != c) (*ei)->clusterize(c);
  }
}

void SilEdge::intersect(SilEdge *e, Camera *vc) {
  if (_v[0] == e->_v[0] || _v[1] == e->_v[0] ||
      _v[0] == e->_v[1] || _v[1] == e->_v[1]) return;

  _intersected.push_back(e);

  double
    x1 =    _iv[0].x(), y1 =    _iv[0].y(),
    x2 =    _iv[1].x(), y2 =    _iv[1].y(),
    x3 = e->_iv[0].x(), y3 = e->_iv[0].y(),
    x4 = e->_iv[1].x(), y4 = e->_iv[1].y();

  double
    det1 = x1*(y2-y3) + x2*(y3-y1) + x3*(y1-y2),
    det2 = x1*(y2-y4) + x2*(y4-y1) + x4*(y1-y2),
    det3 = x3*(y4-y1) + x4*(y1-y3) + x1*(y3-y4),
    det4 = x3*(y4-y2) + x4*(y2-y3) + x2*(y3-y4);

  if ((fabs(det1) < Options::EPSILON) ||
      (fabs(det2) < Options::EPSILON) ||
      (fabs(det3) < Options::EPSILON) ||
      (fabs(det4) < Options::EPSILON)) {

#ifndef NDEBUG
    cerr << "WARNING : potential intersection error" << endl;
#endif

    return;
  }

  if ((det1*det2 >= 0) || (det3*det4 >= 0)) return ;

  double t1 = det3/(det3-det4), t2 = det1/(det1-det2);

  double depth1 = (1-t1)*_iv[0].z() + t1*_iv[1].z();
  double depth2 = (1-t2)*e->_iv[0].z() + t2*e->_iv[1].z();

  Intersection *si;
  if (depth1 < depth2) {
    si = new Intersection(this, e, t2, vc);
    e->_intersection.push_back(si);
  } else if (depth1 > depth2) {
    si = new Intersection(e, this, t1, vc);
    _intersection.push_back(si);
  }
}

void SilEdge::sortIntersection(bool fromhead) {
  Intersection *si;
  if (fromhead) {
    for (int i = _intersection.size() - 1; i > 0; i--) {
      for (int j = 0; j < i; j++) {
	if (_intersection[i]->p() < _intersection[j]->p()) {
	  si = _intersection[i];
	  _intersection[i] = _intersection[j];
	  _intersection[j] = si;
	}
      }
    }
  } else {
    for (int i = _intersection.size() - 1; i > 0; i--) {
      for (int j = 0; j < i; j++) {
	if (_intersection[i]->p() > _intersection[j]->p()) {
	  si = _intersection[i];
	  _intersection[i] = _intersection[j];
	  _intersection[j] = si;
	}
      }
    }
  }
}

int SilEdge::blockingFaces(Vertex *center, const cvec3d & viewpos) {
  int edge_count = 0;
  Vertex *start;

  TMesh::FaceRingType fr;
  cvec3d edgedir;
  if (center == v(0)) {         // center is tri(0)->headVert(eno(0))
    start = v(1);
    fr.collectRing(tri(0), eno(0));
    edgedir = v(1)->limitPos() - v(0)->limitPos();
  } else if (center == v(1)) {  // center is tri(0)->tailVert(eno(0))
    start = v(0);
    fr.collectRing(tri(0), - eno(0));
    edgedir = v(0)->limitPos() - v(1)->limitPos();
  } else die();

  // assert(fr.isClosed());

  cvec3d viewdir = viewpos - center->limitPos();
  cvec3d norm = viewdir.cross(edgedir).dir();
  cvec3d side = norm.cross(viewdir).dir();
  assert(side.dot(edgedir) > 0);
  cvec3d front = edgedir.cross(norm).dir();
  assert(front.dot(viewdir) > 0);

  int fsz = fr.noFace();

  if (fr.isClosed()) {
    cvec3d p1vec = fr.vert(1)->limitPos() - center->limitPos(), p2vec, pvec;
    double t1= norm.dot(p1vec), t2, t;
    bool b1 = (t1 > 0), b2;

    for (int i=2; i<fsz; i++) {
      p2vec = p1vec;
      b2 = b1;
      t2 = t1;
      p1vec = fr.vert(i)->limitPos() - center->limitPos();
      b1 = (t1 = norm.dot(p1vec)) > 0;

      if (b1 == b2) continue;

      t = t1 / (t1 - t2);
      assert( t >=0 && t <= 1);
      pvec = (1-t)*p1vec + t*p2vec;
      if ((side.dot(pvec) > 0) && (front.dot(pvec) > 0)) edge_count++;
    }
  } else {
    cvec3d p1vec = fr.vert(0)->limitPos() - center->limitPos(), p2vec, pvec;
    double t1= norm.dot(p1vec), t2, t;
    bool b1 = (t1 > 0), b2;

    for (int i=1; i<=fsz; i++) {
      p2vec = p1vec;
      b2 = b1;
      t2 = t1;
      p1vec = fr.vert(i)->limitPos() - center->limitPos();
      b1 = (t1 = norm.dot(p1vec)) > 0;

      if ((fr.vert(i-1) == start) ||
	  (fr.vert(i) == start) ||
	  (b1 == b2)) continue;

      t = t1 / (t1 - t2);
      assert( t >=0 && t <= 1);
      pvec = (1-t)*p1vec + t*p2vec;
      if ((side.dot(pvec) > 0) && (front.dot(pvec) > 0)) edge_count++;
    }
  }
  return edge_count;
}

Intersection::Intersection(SilEdge *se, SilEdge *se1, double par, Camera *vc)
  : _e(se), _p(par) {
  cvec3d pt = vc->project(se->tri(0)->headVert
			  (se->tri(0)->nextEno
			   (se->eno(0)))->limitPos());
  cvec3d lndir = se->iv(1) - se->iv(0);
  cvec3d lnnormal(-lndir.y(), lndir.x(), 0);
  double d1 = lnnormal.dot(pt - se->iv(0));
  if ((d1 == 0) && !(se->isBoundary())) {
    pt = vc->project(se->tri(1)->headVert
		     (se->tri(1)->nextEno
		      (se->eno(1)))->limitPos());
    d1 = lnnormal.dot(pt - se->iv(0));
  }
  bool b1 = d1 > 0;
  double d2 = lnnormal.dot(se1->iv(0) - se->iv(0));
  if (d2 == 0) {
    d2 = - lnnormal.dot(se1->iv(1) - se->iv(0));
  }
  bool b2 = d2 > 0;
  _headOccluded = (b1 == b2);
}

void Chain::setHead(SilEdge *e, double p) {
  assert((_edges.size() == 0) && (_points.size() == 0));
  _edges.push_back(e); 
  _head = p;
  _points.push_back((1-p)*e->iv(0) + p*e->iv(1));
}

void Chain::insert(SilEdge *e, bool fromhead) {
  assert((_edges.size() > 0) && (_points.size() > 0));
  _edges.push_back(e);
  if (fromhead) _points.push_back(e->iv(0));
  else _points.push_back(e->iv(1));
}

void Chain::setTail(SilEdge *e, double p) {
  assert((_edges.size() > 0) && (_points.size() > 0));
  _tail = p;
  _points.push_back((1-p)*e->iv(0) + p*e->iv(1));
}

Grid2D::Grid2D(int s) : _size(s), _xstep(0), _ystep(0),
  _min(0), _max(0) {
  _grid = new vector<SilEdge*>*[_size];
  for (int ind = 0; ind < _size; ind++)
    _grid[ind] = new vector<SilEdge*>[_size];  
}

Grid2D::~Grid2D() {
  clear();
  for (int ind = 0; ind < _size; ind++)
    delete[] _grid[ind];
  delete[] _grid;
  _grid = 0;
}

void Grid2D::placeIntoGrid(vector<SilEdge*>& edges) {
  if (edges.size() == 0) return;

  // compute bounding box
  double minx = DBL_MAX, miny = DBL_MAX, maxx = -DBL_MAX, maxy = -DBL_MAX;
  int it, esz = edges.size();

  for (it = 0; it < esz; it++) {
    SilEdge *edge = edges[it];
    cvec3d p1 = edge->iv(0);
    cvec3d p2 = edge->iv(1);
    if (minx > p1.x()) minx = p1.x();
    else if (maxx < p1.x()) maxx = p1.x();
    if (miny > p1.y()) miny = p1.y();
    else if (maxy < p1.y()) maxy = p1.y();
    if (minx > p2.x()) minx = p2.x();
    else if (maxx < p2.x()) maxx = p2.x();
    if (miny > p2.y()) miny = p2.y();
    else if (maxy < p2.y()) maxy = p2.y();
  }

  maxx += 1; maxy += 1;
  minx -= 1; miny -= 1;
  _xstep = (maxx - minx) / _size;
  _ystep = (maxy - miny) / _size;
  _min = cvec3d( minx, miny, 0 );
  _max = cvec3d( maxx, maxy, 0 );

  // put edges into grid, by scan conversion
  for (it = 0; it < esz; it++) {
    SilEdge *edge = edges[it];
    cvec3d p1 = edge->iv(0);
    cvec3d p2 = edge->iv(1);
    cvec3d dir = p2 - p1;
    int idir[] = {dir.x() > 0 ? 1 : (dir.x() == 0 ? 0 : -1),
		  dir.y() > 0 ? 1 : (dir.y() == 0 ? 0 : -1)};

    // first cell
    int ix = (int)floor((p1.x() - _min.x())/_xstep);
    int iy = (int)floor((p1.y() - _min.y())/_ystep);
    double tx = 1.0, ty = 1.0, dist = 0.0;

    while ((dist < 1.0) && 
	   (ix >= 0) && (ix < _size) && 
	   (iy >= 0) && (iy < _size)) {
      _grid[ix][iy].push_back(edge);
      // next cell
      if (idir[0] != 0) {
	tx = (_min.x() + (ix+(idir[0]+1)/2)*_xstep - p1.x()) / dir.x();
      }
      if (idir[1] != 0) {
	ty = (_min.y() + (iy+(idir[1]+1)/2)*_ystep - p1.y()) / dir.y();
      }
      if (tx == ty) {
	dist = tx;
	int xx = ix + idir[0], yy = iy + idir[1];
	if ((xx >= 0) && (xx < _size)) {
	  _grid[xx][iy].push_back(edge);
	}
	if ((yy >= 0) && (yy < _size)) {
	  _grid[ix][yy].push_back(edge);
	}
	ix += idir[0];
	iy += idir[1];
      } else if (tx < ty) {
	dist = tx;
	ix += idir[0];
      } else {
	dist = ty;
	iy += idir[1];
      }
    } /* while */
    // last cell
    if ((ix >= 0) && (ix < _size) && 
	(iy >= 0) && (iy < _size)) {
      _grid[ix][iy].push_back(edge);
    }
  } /* for each edge */
}

void Grid2D::clear() {
  for (int i = 0; i < _size; i++)
    for (int j = 0; j < _size; j++)
      _grid[i][j].clear();
}

Silhouettes::~Silhouettes() {
  clear();

  if (_grid) delete _grid;
  _grid = 0;
}

void Silhouettes::insertEdge(DualEdge *e) {
  Tri *t0 = e->tri(0);
  Tri *t1 = e->tri(1);

  // check if the edge is already here
  if (siltripairs.find( tripair(t0, t1) ) != siltripairs.end()) return;
  siltripairs.insert( tripair(t0, t1) );

  SilEdge *se = new SilEdge(e, _camera);
  _edges.push_back(se);

  t0->headVert(e->eno(0))->insertSilEdge(se);
  t0->tailVert(e->eno(0))->insertSilEdge(se);
}

void Silhouettes::insertEdge(BoundaryEdge *e) {
  SilEdge *se = new SilEdge(e, _camera);
  _edges.push_back(se);

  e->v1()->insertSilEdge(se, true);
  e->v2()->insertSilEdge(se, true);
}

void Silhouettes::clear() {
  // clear memory
  SilEdge::clear();
  Intersection::clear();
  Chain::clear();

  // clear grid
  if (_grid) _grid->clear();

  // clear rays
  if (_meshgrid) _meshgrid->clearRays();

  // clear chains and edges
  int sz = _chains.size();
  for (int i=0; i<sz; i++) delete _chains[i];
  _chains.clear();
  sz = _edges.size();
  for (int j=0; j<sz; j++) delete _edges[j];
  _edges.clear();

  siltripairs.clear();
}

void Silhouettes::createGrid(int size) {
  if (_grid) delete _grid;
  _grid = new Grid2D(size);
}

void Silhouettes::clusterize() {
  int cl = 1;
  for (vector<SilEdge*>::iterator ei = _edges.begin();
       ei != _edges.end(); ei++) {
    if ((*ei)->cluster() < 0) (*ei)->clusterize(cl++);
  }
}

void Silhouettes::findIntersections() {
  int sz = _grid->size();
  int ix, iy;
  vector<SilEdge*> *g;
  int ss, e1, e2;
  SilEdge *se1, *se2;

  _grid->placeIntoGrid(_edges);

  for (ix = 0; ix < sz; ix++) {
    g = (*_grid)[ix];
    for (iy = 0; iy < sz; iy++) {
      ss = g[iy].size();
      for (e1 = 0; e1 < ss-1; e1++) {
	for (e2 = e1+1; e2 < ss; e2++) {
	  if (g[iy][e1] < g[iy][e2]) {
	    se1 = g[iy][e1];
	    se2 = g[iy][e2];
	  } else {
	    se1 = g[iy][e2];
	    se2 = g[iy][e1];
	  }
	  if (!se1->intersected(se2)) {
	    se1->intersect(se2, _camera);
	  }
	}
      }
    }
  }

}

// cusp vertices definition
class Cusp {
public:
  Vertex *v;
  int relativeQI;
  Cusp(Vertex *vv, int q) : v(vv), relativeQI(q) { };
};

void Silhouettes::createChains() {
  if (_edges.empty()) return;

  clusterize();

  findIntersections();

  vector<Chain*> curCluster;
  Chain *chain, *minChain;
  int ei = 0, edgesize = _edges.size();
  SilEdge *edge = _edges[ei], *pre_edge;
  Vertex *start_vertex = edge->v(0);
  double phead = 0.0, ptail = 1.0;
  bool fromhead = true, pre_fromhead;
  int curQI = 0, minQI = 0, bf1, bf2;

  vector<Cusp> cusps;

  cvec3d viewpos = _camera->viewWorldPosition();

  // if start vertex is a cusp
  if (start_vertex->silEdges().size() > 1)
    if (start_vertex->cusp() ||
	(start_vertex->silEdges()[0]->facing() !=
	 start_vertex->silEdges()[1]->facing())) {
      int q = - edge->blockingFaces(start_vertex, viewpos);
      cusps.push_back(Cusp(start_vertex, q));
      start_vertex = 0;
    }

  chain = new Chain();
  _chains.push_back(chain);
  curCluster.push_back(chain);
  minChain = chain;
  chain->setQI(curQI);
  chain->setHead(edge, phead);

  // put all edges on chains
  while (ei < edgesize) {
    edge->setOnChain(true);
    edge->v(0)->eraseSilEdge(edge);
    edge->v(1)->eraseSilEdge(edge);

    // cut edges by intersections
    int nInter = edge->nInter();
    if (nInter > 0) {
      edge->sortIntersection(fromhead);
      for (int i = 0; i < nInter; i++) {
	ptail = edge->intersection(i)->p();
	chain->setTail(edge, ptail);
	chain = new Chain();
	_chains.push_back(chain);
	curCluster.push_back(chain);

	if (fromhead == edge->intersection(i)->headOccluded()) {
	  if (edge->intersection(i)->e()->isBoundary()) curQI -= 1;
	  else curQI -= 2;
	  if (minQI > curQI) {
	    minQI = curQI;
	    minChain = chain;
	  }
	} else {
	  if (edge->intersection(i)->e()->isBoundary()) curQI += 1;
	  else curQI += 2;
	}
	chain->setQI(curQI);
	chain->setHead(edge, ptail);
      }
    } // intersection finished

    // next edge
    Vertex* start = edge->v( fromhead ? 1 : 0 );
    // there are edges sharing the start vertex with this edge
    if (!start->silEdges().empty()) {
      pre_edge = edge;
      pre_fromhead = fromhead;
      edge = start->silEdges()[0];
      if (edge->v(0) == start) {
	fromhead = true;
	phead = 0.0;
      } else {
	fromhead = false;
	phead = 1.0;
      }

      // chain should be cut at cusps
      if ((edge->facing() != pre_edge->facing()) || start->cusp()) {
	chain->setTail(pre_edge, pre_fromhead ? 1.0 : 0.0);
	chain = new Chain();
	_chains.push_back(chain);
	curCluster.push_back(chain);

	bf1 = pre_edge->blockingFaces(start, viewpos);
	if (start->silEdges().size() > 1) {
	  cusps.push_back(Cusp(start, curQI - bf1));
	}
	bf2 = edge->blockingFaces(start, viewpos);

	curQI += bf2 - bf1;
	if (minQI > curQI) {
	  minQI = curQI;
	  minChain = chain;
	}

	chain->setQI(curQI);
	chain->setHead(edge, phead);
      } else {
	chain->insert(edge, fromhead);
      }
    }
    // no more edges at this vertex
    else {
      // end this chain
      chain->setTail(edge, fromhead ? 1.0 : 0.0);

      // check if start_vertex has any edges left
      if (start_vertex) {
	if (!start_vertex->silEdges().empty()) {
	  chain = new Chain();
	  _chains.push_back(chain);
	  curCluster.push_back(chain);
	  edge = start_vertex->silEdges()[0];
	  if (start_vertex == edge->v(0)) {
	    fromhead = true; phead = 0;
	  } else {
	    fromhead = false; phead = 1;
	  }
	  curQI = 0;
	  chain->setQI(curQI);
	  chain->setHead(edge, phead);

	  start_vertex = 0;
	  continue;
	}
      }

      // deal with cusps
      if (!cusps.empty()) {
	vector<Cusp>::iterator ci = cusps.begin();
	while (ci != cusps.end()) {
	  if ((*ci).v->silEdges().empty()) ci++;
	  else break;
	}
	if (ci == cusps.end()) {
	  cusps.clear();
	}
	// there are cusps with edges unfinished
	else {
	  chain = new Chain();
	  _chains.push_back(chain);
	  curCluster.push_back(chain);

	  edge = (*ci).v->silEdges()[0];
	  if ((*ci).v == edge->v(0)) {
	    fromhead = true; phead = 0;
	  } else {
	    fromhead = false; phead = 1;
	  }
	  curQI = (*ci).relativeQI + edge->blockingFaces((*ci).v, viewpos);
	  if (minQI > curQI) {
	    minQI = curQI;
	    minChain = chain;
	  }
	  chain->setQI(curQI);
	  chain->setHead(edge, phead);

	  if (ci != cusps.begin())
	    cusps.erase(cusps.begin(), ci);
	  continue;
	}
      } // cusps finished

      // finishing this cluster
      int clsz = curCluster.size();
      if (minChain->size() < 6) {
	for (int it = 0; it < clsz; it++) {
	  if (curCluster[it]->QI() == minQI &&
	      curCluster[it]->size() > minChain->size()) {
	    minChain = curCluster[it];
	    if (minChain->size() >= 6) break;
	  }
	}
      }

      // do raycasting for initial visibility
      {
	// choose ray distination
	cvec3d rayEnd;
	int edgesize = minChain->size();
	if (edgesize > 2)
	  rayEnd = minChain->edge(edgesize/2)->v(0)->limitPos();
	else if (edgesize == 2)
	  rayEnd = _camera->unproject(minChain->points()[1]);
	else
	  rayEnd = _camera->unproject(0.5*(minChain->points()[0] +
					   minChain->points()[1]));

	if (_meshgrid->intersectRay(viewpos, rayEnd))
	  minQI--;
      } // ray visibility

      if (minQI != 0) {
	for (int it = 0; it < clsz; it++) {
	  curCluster[it]->modifyQI( - minQI );
	}
      }
      curQI = minQI = 0;
      curCluster.clear();
      cusps.clear();
      // cluster finished

      // next edge for next cluster
      while (ei < edgesize) {
	if (_edges[ei]->onChain()) ei++;
        else break;
      }

      if (ei < edgesize) {
	edge = _edges[ei];
	fromhead = true;
	phead = 0.0;
	start_vertex = edge->v(0);

	// if start vertex is a cusp
	if (start_vertex->silEdges().size() > 1)
	  if (start_vertex->cusp() ||
	      (start_vertex->silEdges()[0]->facing() !=
	       start_vertex->silEdges()[1]->facing())) {
	    int q = - edge->blockingFaces(start_vertex, viewpos);
	    cusps.push_back(Cusp(start_vertex, q));
	    start_vertex = 0;
	  }

	chain = new Chain();
	_chains.push_back(chain);
	curCluster.push_back(chain);
	minChain = chain;
	chain->setQI( curQI );
	chain->setHead(edge, phead);
      } // next edge picked
    }
  }

}

