// -*- Mode: c++ -*-
// $Id: camera.cpp,v 1.7 2000/07/21 01:33:17 jianbo Exp $
// $Source: /tools/cvs-repos/illustrate/sub/camera.cpp,v $

/* Subdivide V2.0
   Copyright (C) 2000 Henning Biermann, Denis Zorin, NYU

This file is part of Subdivide.

Subdivide is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Subdivide is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Subdivide; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include <string.h>
#include "camera.h"

// storage for current matrices
GLdouble Camera::_proj[16];
GLdouble Camera::_view[16];
GLint Camera::_vp[4];

// ******************** Camera memeber functions ***********************

void Camera::identityModel() {
  _ballModel.setIdentity();
}

Camera::Camera(const Camera & c) :
  _projectionMatrix(c._projectionMatrix),
  _modelviewMatrix(c._modelviewMatrix),
  _trans(c._trans),
  _ballModel(c._ballModel),
  _znear(c._znear), 
  _zfar(c._zfar),
  _fovy(c._fovy), 
  _aspect(c._aspect),
  _wincenterx(c._wincenterx),
  _wincentery(c._wincentery),
  _winscale(c._winscale)
{

  _viewport[0] = c._viewport[0];
  _viewport[1] = c._viewport[1];
  _viewport[2] = c._viewport[2];
  _viewport[3] = c._viewport[3];
}

const Camera & Camera::operator=(const Camera & c) {
  _trans = c._trans;
  _fovy = c._fovy; _znear = c._znear; _zfar = c._zfar;
  _wincenterx =  c._wincenterx; _wincentery =  c._wincentery;
  _winscale = c._winscale;
  _ballModel = c._ballModel;
  _modelviewMatrix = c._modelviewMatrix;
  _projectionMatrix = c._projectionMatrix;

  _aspect = c._aspect;
  _viewport[0] = c._viewport[0];
  _viewport[1] = c._viewport[1];
  _viewport[2] = c._viewport[2];
  _viewport[3] = c._viewport[3];
  return *this;
}

void Camera::copyModelOnly(const Camera & c) {
  _trans = c._trans;
  _ballModel = c._ballModel;
  _modelviewMatrix = c._modelviewMatrix;
}

void Camera::reset() {
  _trans = 0;
  _fovy = 34.0;
  _znear = 0.1;
  _zfar = 100.0;
  _winscale = 1.0;
  _wincenterx = 0.0;
  _wincentery = 0.0;
  _aspect = 1.0;
  identityModel();

  computeModelview();
  computeProjection();
}

void Camera::setFovy(double fovy_degrees) {
  _fovy = fovy_degrees;
  computeProjection();
}

void Camera::setPerspectiveParams(double fovy_degrees, double znear,double
                                  zfar) {
    _fovy = fovy_degrees;  
    _znear = znear;
    _zfar = zfar;
    computeProjection();
  }


// ******************** OpenGL matrices ********************

// TODO: these need not use OpenGL
void Camera::computeModelview() {
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(_trans.x(), _trans.y(), _trans.z());

  glMultMatrixd( (double*)(_ballModel));
  glGetDoublev(GL_MODELVIEW_MATRIX,(double*)(_modelviewMatrix));
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

}

// TODO: these need not use OpenGL
void Camera::computeProjection() {

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  double ysize = tan(0.5*_fovy*M_PI/180.0)*_znear;
  double xsize = _aspect*ysize;
  double left   = _wincenterx -  xsize*_winscale; 
  double right  = _wincenterx +  xsize*_winscale; 
  double bottom = _wincentery -  ysize*_winscale; 
  double top    = _wincentery +  ysize*_winscale; 

  glFrustum(left, right, bottom, top, _znear, _zfar);

  glGetDoublev(GL_PROJECTION_MATRIX,(double*)(_projectionMatrix));
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);

}

void Camera::loadMatrices()const {
  glMatrixMode(GL_PROJECTION);
  glLoadMatrixd((const double*)(_projectionMatrix));
  glMatrixMode(GL_MODELVIEW);
  glLoadMatrixd((const double*)(_modelviewMatrix));
}
    

// ******************** project/unproject ********************

// project a point from world space to screen space
cvec3d Camera::project(const cvec3d&  worldPoint) const {
  GLdouble winx, winy, winz;

  gluProject(worldPoint.x(),worldPoint.y(),worldPoint.z(),
	 (const double*)(_modelviewMatrix),
             (const double*)(_projectionMatrix),
             _viewport,
             &winx,&winy,&winz);

  return cvec3d(winx,winy,winz);
}

// compute the world space coordinate for an image coordinate
cvec3d Camera::unproject(const cvec3d& imagePoint) const {
  GLdouble objx, objy, objz;

  gluUnProject(imagePoint.x(),imagePoint.y(),imagePoint.z(),
               (const double*)(_modelviewMatrix),
               (const double*)(_projectionMatrix)
               ,_viewport,
	     &objx,&objy,&objz);

  return cvec3d(objx,objy,objz);
}

// compute the object space coordinate for an image coordinate
cvec3d Camera::unproject(const HMatrix& model, const cvec3d& imagePoint) const {
  GLdouble objx, objy, objz;

  gluUnProject(imagePoint.x(),imagePoint.y(),imagePoint.z(),
               (const double*)(_modelviewMatrix * model),
               (const double*)(_projectionMatrix)
               ,_viewport,
	     &objx,&objy,&objz);

  return cvec3d(objx,objy,objz);
}


// ******************** camera modification ********************

void Camera::translate(const cvec3d & t) { _trans += t;}

void Camera::translateWindow(double dx, double dy) { 
 _wincentery -= 2.0*dy*tan(0.5*_fovy*M_PI/180.0)*_znear*_winscale;
 _wincenterx -= 2.0*dx*tan(0.5*_fovy*M_PI/180.0)*_aspect*_znear*_winscale;
 computeProjection();
}

void Camera::scaleWindow( double sfactor) { 
  _winscale = _winscale*pow(2.0, sfactor);
  computeProjection();
}


// ******************** input/output  ********************

ostream& operator<<(ostream& os, const Camera& camera ) {
  os.precision(16);
  os << "TRANSFORMATIONS:\n";
  os << "ModelviewMatrix: "  << camera._modelviewMatrix  << endl;
  os << "ProjectionMatrix: " << camera._projectionMatrix << endl;

  os << "\nPARAMETERS:\n";
  os << "ballModel: " << camera._ballModel << endl;
  os << "trans:  "    << camera._trans  << endl;
  os << "aspect: "    << camera._aspect << endl;
  os << "znear:  "    << camera._znear  <<  endl; 
  os << "zfar:   "    << camera._zfar   << endl;
  os << "fovy:   "    << camera._fovy   << endl;
  return os;
}

istream& operator>>(istream& is, Camera& camera) { 
  char buf[256] ;
  is >> buf ;
  assert(strcmp(buf, "TRANSFORMATIONS:") == 0);
  is >> buf ;
  assert(strcmp(buf, "ModelviewMatrix:") == 0);
  is >> camera._modelviewMatrix ;
  is >> buf ;
  assert(strcmp(buf, "ProjectionMatrix:") == 0);
  is >> camera._projectionMatrix ;
  is >> buf ;
  assert(strcmp(buf, "PARAMETERS:") == 0);
  is >> buf ;
  assert(strcmp(buf, "ballModel:") == 0);
  is >> camera._ballModel ;
  is >> buf ;
  assert(strcmp(buf, "trans:") == 0);
  is >> camera._trans ;
  is >> buf ;
  assert(strcmp(buf, "aspect:") == 0);
  is >> camera._aspect ;
  is >> buf ;
  assert(strcmp(buf, "znear:") == 0);
  is >> camera._znear ;
  is >> buf ;
  assert(strcmp(buf, "zfar:") == 0);
  is >> camera._zfar ;
  is >> buf ;
  assert(strcmp(buf, "fovy:") == 0);
  is >> camera._fovy ;
  return is;
}

// ********************** current matrix project *********

/*
void Camera::getMatricesC()const {
  glGetDoublev(GL_MODELVIEW_MATRIX,(double*)_view);
  glGetDoublev(GL_PROJECTION_MATRIX,(double*)_proj);
  glGetIntegerv(GL_VIEWPORT,_vp);
}

cvec3d Camera::projectC(const cvec3d&  worldPoint) const {
  GLdouble winx, winy, winz;
  gluProject(worldPoint.x(),worldPoint.y(),worldPoint.z(),
	     _view,_proj,_vp,&winx,&winy,&winz);
  return cvec3d(winx,winy,winz);
}

cvec3d Camera::unprojectC(const cvec3d& screenPoint) const {
  GLdouble wx, wy, wz;
  gluUnProject(screenPoint.x(),screenPoint.y(),screenPoint.z(),
	       _view, _proj, _vp, &wx, &wy, &wz);
  return cvec3d(wx,wy,wz);
}
*/

